@@ -1,6 +1,6 @@
 /*
- * Copyright (C) 2005  Stig Venaas <venaas@uninett.no>
- * $Id: ssmpngcl.c,v 1.4 2005/11/29 16:27:26 sv Exp $
+ * Copyright (C) 2005, 2006  Stig Venaas <venaas@uninett.no>
+ * $Id:$
  *
  * Contributions:
  * Solaris support by Alexander Gall <gall@switch.ch>
@@ -14,6 +14,44 @@
 
 #include "ssmping.h"
 
+#ifdef WIN32
+#include "mswsock.h"
+
+/* further definitions missing from my mswsock.h */
+#ifndef WSAID_WSARECVMSG
+/* http://cvs.winehq.org/cvsweb/wine/include/mswsock.h */
+#define WSAID_WSARECVMSG {0xf689d7c8,0x6f1f,0x436b,{0x8a,0x53,0xe5,0x4f,0xe3,0x51,0xc3,0x22}}
+#endif
+#ifndef CMSG_FIRSTHDR
+#define CMSG_FIRSTHDR(pwsamsg) (pwsamsg)->Control.len >= sizeof(WSACMSGHDR) ? (WSACMSGHDR *)((pwsamsg)->Control.buf) : NULL
+#endif
+#ifndef CMSG_NXTHDR
+#define CMSG_NXTHDR(pwsamsg, cmsg) my_cmsg_nxthdr(pwsamsg, cmsg)
+WSACMSGHDR *my_cmsg_nxthdr(WSAMSG *mhdr, WSACMSGHDR *cmsg) {
+    WSACMSGHDR *cmsgh = (WSACMSGHDR *)(((char *)cmsg) + cmsg->cmsg_len);
+    /* should also do alignment above */
+    return (char *)(cmsgh + 1) <= (char *)mhdr->Control.buf + mhdr->Control.len ? cmsgh : NULL;
+}
+#endif
+#ifndef MSG_MCAST
+#define MSG_MCAST 2048
+#endif
+#ifndef IPV6_HOPLIMIT
+#define IPV6_HOPLIMIT 21
+#endif
+#ifndef CMSG_DATA
+#define MY_CMSG_LEN(len) (len)+sizeof(WSACMSGHDR)
+#define CMSG_DATA(cmsg) ((char *)(cmsg)+sizeof(WSACMSGHDR))
+#endif
+#ifndef _CMSG_DATA_ALIGN
+#define _CMSG_DATA_ALIGN(len) (((len) + sizeof (size_t) - 1) & (size_t) ~(sizeof (size_t) - 1))
+#endif
+#endif
+
+#ifndef MY_CMSG_LEN
+#define MY_CMSG_LEN CMSG_LEN
+#endif
+
 /* llsqrt() taken from Linux's iputils package */
 static long llsqrt(unsigned long long a) {
     long long prev = ~((unsigned long long) 1 << 63);
@@ -108,7 +146,7 @@ int addr_equal(struct sockaddr *a, struct sockaddr *b) {
     }
 }
 
-int dupcheck(int s, unsigned int t) {
+int dupcheck(char s, unsigned int t) {
     static int first = 1;
     static unsigned int recent[2][100];
     int i = s % 2;
@@ -140,17 +178,15 @@ void prep_sock(int family, int s) {
 #endif
 
     switch (family) {
-#ifndef WIN32	
     case AF_INET6:
 #ifdef IPV6_RECVHOPLIMIT
 	if (setsockopt(s, IPPROTO_IPV6, IPV6_RECVHOPLIMIT, &on, sizeof(on)) == -1)
 	    errx("setsockopt IPV6_RECVHOPLIMIT");
 #else
-	if (setsockopt(s, IPPROTO_IPV6, IPV6_HOPLIMIT, &on, sizeof(on)) == -1)
+	if (setsockopt(s, IPPROTO_IPV6, IPV6_HOPLIMIT, (char *)&on, sizeof(on)) == -1)
 	    errx("setsockopt IPV6_HOPLIMIT");
 #endif
 	break;
-#endif
     case AF_INET:
 #ifdef IP_RECVTTL
 	if (setsockopt(s, IPPROTO_IP, IP_RECVTTL, &on, sizeof(on)) == -1)
@@ -158,7 +194,7 @@ void prep_sock(int family, int s) {
 #else
 	if (setsockopt(s, IPPROTO_IP, IP_TTL, (char *)&on, sizeof(on)) == -1)
 	    errx("setsockopt IP_TTL");
-#endif	
+#endif
 	break;
     }
 }
@@ -182,22 +218,61 @@ void findsrc(struct sockaddr *src, struct sockaddr *dst) {
     close(s);
 }
 
+/* mc is a flag saying if multicast or unicast, only used on WIN32 */
 int recvfromhopstime(int s, void *buf, size_t len, int flags, struct sockaddr *from,
-		     socklen_t *fromlen, int32_t *hops, struct timeval *tstamp) {
-#ifndef WIN32
+		     socklen_t *fromlen, int32_t *hops, struct timeval *tstamp, char *mc) {
+#ifdef WIN32
+    GUID WSARecvMsg_GUID = WSAID_WSARECVMSG;
+    long ioctlcount;
+    static int (*WSARecvMsg)() = NULL;
+    WSAMSG msgh;
+    WSABUF iovec;
+    WSACMSGHDR *cmsgh;
+#else    
     struct msghdr msgh;
     struct iovec iovec;
-    char control[64];
     struct cmsghdr *cmsgh;
-    int cnt;
 #endif
+    char control[1024];
+    int cnt;
 
+#ifdef WIN32
+    if (!WSARecvMsg) {
+	if (WSAIoctl(s, SIO_GET_EXTENSION_FUNCTION_POINTER, &WSARecvMsg_GUID, sizeof(WSARecvMsg_GUID), &WSARecvMsg, sizeof(WSARecvMsg), &ioctlcount, NULL, NULL)) {
+	    fprintf(stderr, "WSAIoctl failed with code %d\n", WSAGetLastError());
+	    return -1;
+	}
+    }
+#endif
+    
     if (hops)
 	*hops = -1;
     if (tstamp)
 	memset(tstamp, 0, sizeof(struct timeval));
     
-#ifndef WIN32
+#ifdef WIN32
+    memset(&msgh, 0, sizeof(msgh));
+    msgh.lpBuffers = &iovec;
+    msgh.dwBufferCount = 1;
+    msgh.Control.buf = control;
+    msgh.Control.len = sizeof(control);
+    memset(control, 0, sizeof(control));
+    msgh.name = from;
+    msgh.namelen = *fromlen;
+    msgh.dwFlags = 0;
+    memset(&iovec, 0, sizeof(iovec));
+    iovec.buf = buf;
+    iovec.len = len;
+
+    if (WSARecvMsg(s, &msgh, &cnt, NULL, NULL)) {
+	    fprintf(stderr, "WSARecvMsg failed with code %d\n", WSAGetLastError());
+	    return -1;
+    }
+
+    if (mc)
+	*mc = (msgh.dwFlags & MSG_MCAST) == MSG_MCAST;
+
+#else    
     memset(&msgh, 0, sizeof(struct msghdr));
     msgh.msg_iov = &iovec;
     msgh.msg_iovlen = 1;
@@ -210,20 +285,33 @@ int recvfromhopstime(int s, void *buf, size_t len, int flags, struct sockaddr *f
     iovec.iov_len = len;
 
     cnt = recvmsg(s, &msgh, 0);
+#endif
+    
     if (cnt < 1 || (!hops && !tstamp))
 	return cnt;
+
+#if 0
+    printf("flags=%d\n", msgh.dwFlags);
+    {
+	int i;
+	for (i = 0; i < 16; i++)
+	    printf("%hhd\n", control[i]);
+    }
+
+    printf("control length = %d\n", msgh.Control.len);
+#endif
     
     for (cmsgh = CMSG_FIRSTHDR(&msgh); cmsgh; cmsgh = CMSG_NXTHDR(&msgh, cmsgh))
 	switch (cmsgh->cmsg_level) {
 #ifdef SCM_TIMESTAMP
 	case SOL_SOCKET:
 	    if (cmsgh->cmsg_type == SCM_TIMESTAMP &&
-		cmsgh->cmsg_len >= CMSG_LEN(sizeof(struct timeval)))
+		cmsgh->cmsg_len >= MY_CMSG_LEN(sizeof(struct timeval)))
 		*tstamp = *(struct timeval *)CMSG_DATA(cmsgh);
 	    break;
 #endif
 	case IPPROTO_IP:
-	    if (cmsgh->cmsg_type == IP_TTL && cmsgh->cmsg_len >= CMSG_LEN(sizeof(int32_t)))
+	    if (cmsgh->cmsg_type == IP_TTL && cmsgh->cmsg_len >= MY_CMSG_LEN(sizeof(int32_t)))
 		*hops = *(int *)CMSG_DATA(cmsgh);
 #ifdef IP_RECVTTL
 	    /* Only found Solaris 9 to use IP_RECVTTL so far */
@@ -232,17 +320,14 @@ int recvfromhopstime(int s, void *buf, size_t len, int flags, struct sockaddr *f
 #endif
 	    break;
 	case IPPROTO_IPV6:
-	    if (cmsgh->cmsg_type == IPV6_HOPLIMIT && cmsgh->cmsg_len >= CMSG_LEN(sizeof(int)))
+	    if (cmsgh->cmsg_type == IPV6_HOPLIMIT && cmsgh->cmsg_len >= MY_CMSG_LEN(sizeof(int)))
 		*hops = *(int *)CMSG_DATA(cmsgh);
 	    break;
 	}
     return cnt;
-#else
-    return recvfrom(s, buf, len, flags, from, fromlen);
-#endif    
 }
 
-size_t initsendbuf(char *buf, size_t buflen, pid_t pid, struct sockaddr_storage *group,
+size_t initsendbuf(char *buf, size_t buflen, pid_t pid, int ver, uint16_t size, struct sockaddr_storage *group,
 		   void **seq, void **timestamp) {
     char *p = buf;
     uint32_t int32[2];
@@ -273,6 +358,20 @@ size_t initsendbuf(char *buf, size_t buflen, pid_t pid, struct sockaddr_storage
     *timestamp = p;
     p += 8;
 
+    if (ver) {
+	if (p + tlvspace(0) > buf + buflen)
+	    errx("Send buffer too small");
+	p = tlvadd(p, SSMPING_RQVER, 0, NULL);
+    }
+
+    if (size) {
+	uint16_t tsize = htons(size);
+	if (p + tlvspace(2) > buf + buflen)
+	    errx("Send buffer too small");
+	p = tlvadd(p, SSMPING_REPLYSIZE, 2, &tsize);
+	p += 2;
+    }
+    
     switch (((struct sockaddr *)group)->sa_family) {
     case AF_INET:
 	if (p + tlvspace(5) > buf + buflen)
@@ -294,13 +393,12 @@ size_t initsendbuf(char *buf, size_t buflen, pid_t pid, struct sockaddr_storage
     return p - buf;
 }
 
-int parsepacket(char *buf, size_t len, struct ssmpingdata *data) {
+int parsepacket(char *buf, size_t len, char **verstring, struct ssmpingdata *data) {
     uint16_t t, l, tmp;
-    char *p = buf;
-    void *v;
+    char *v, *p = buf;
     uint32_t val;
     int pid = 0, seq = 0, timestamp = 0;
-    
+
     while (p - buf + 4 <= len) {
 	memcpy(&tmp, p, 2);
 	t = ntohs(tmp);
@@ -338,6 +436,17 @@ int parsepacket(char *buf, size_t len, struct ssmpingdata *data) {
 	    data->timestamp.tv_usec = ntohl(val);
 	    timestamp = 1;
 	    break;
+	case SSMPING_VER:
+	    if (!l)
+		return -1;
+	    if (!*verstring) {
+		*verstring = malloc(l);
+		if (*verstring) {
+		    memcpy(*verstring, v, l);
+		    (*verstring)[l] = '\0';
+		}
+	    }
+	    break;
 	}
     }
     if (p - buf != len)
@@ -345,15 +454,38 @@ int parsepacket(char *buf, size_t len, struct ssmpingdata *data) {
     return 0;
 }
 
-int name2addrsock(const char *name, int *family, struct sockaddr_storage *addr) {
-    struct addrinfo hints, *res;
-    int e, s;
+int ismc(struct sockaddr *sa) {
+    switch (sa->sa_family) {
+    case AF_INET: return IN_MULTICAST(ntohl(((struct sockaddr_in *)sa)->sin_addr.s_addr));
+    case AF_INET6: return IN6_IS_ADDR_MULTICAST(&((struct sockaddr_in6 *)sa)->sin6_addr);
+    }
+    return 0;
+}
+
+/* return 0 if ok, -1 on error. creates sockets s1, s2 if not NULL
+ * mcaddr must be mc and ucaddr uc, unless both specified, in which
+ * case one must be uc and the other mc
+ */
+int names2addrsocks(int *s1, int *s2, const char *ucaddr, const char *mcaddr, const char *srv, int *family,
+		     struct sockaddr_storage *ucsa, struct sockaddr_storage *mcsa) {
+    struct addrinfo hints, *res1, *res2;
+    const char *addr;
+    int e;
+
+    addr = mcaddr ? mcaddr : ucaddr;
+    if (!addr)
+	return -1;
+    
+    if (s1)
+	*s1 = -1;
+    if (s2)
+	*s2 = -1;
     
     memset(&hints, 0, sizeof(hints));
     hints.ai_family = *family;
     hints.ai_socktype = SOCK_DGRAM;
 
-    if ((e = getaddrinfo(name, "4321", &hints, &res))) {
+    if ((e = getaddrinfo(addr, srv, &hints, &res1))) {
 #ifdef WIN32
 	err("getaddrinfo failed with error code %d", e);
 #else    
@@ -361,32 +493,104 @@ int name2addrsock(const char *name, int *family, struct sockaddr_storage *addr)
 #endif
 	return -1;
     }
-    for (; res; res = res->ai_next) {
-	s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
-	if (s >= 0) {
-	    *family = res->ai_family;
-	    memcpy(addr, res->ai_addr, res->ai_addrlen);
-	    return s;
+    for (; res1; res1 = res1->ai_next) {
+	if (s1) {
+	    *s1 = socket(res1->ai_family, res1->ai_socktype, res1->ai_protocol);
+	    if (*s1 < 0) {
+		err("socket");
+		continue;
+	    }
 	}
-	err("socket");
+	if (s2) {
+	    *s2 = socket(res1->ai_family, res1->ai_socktype, res1->ai_protocol);
+	    if (*s2 < 0) {
+		if (s1) {
+		    close(*s1);
+		    *s1 = -1;
+		}
+		err("socket");
+		continue;
+	    }
+	}
+	
+	*family = res1->ai_family;
+		
+	if (!ucaddr || !mcaddr) {
+	    if (ismc(res1->ai_addr)) {
+		if (mcaddr) {
+		    memcpy(mcsa, res1->ai_addr, res1->ai_addrlen);
+		    return 0;
+		}
+		err("Argument must be a unicast address");
+		return -1;
+	    }
+	    if (ucaddr) {
+		memcpy(ucsa, res1->ai_addr, res1->ai_addrlen);
+		return 0;
+	    }
+	    err("Argument must be a multicast address");
+	    return -1;
+	}
+    
+	hints.ai_family = *family;
+	if ((e = getaddrinfo(ucaddr, srv, &hints, &res2))) {
+#ifdef WIN32
+	    err("getaddrinfo failed with error code %d", e);
+#else    
+	    err("getaddrinfo failed: %s", gai_strerror(e));
+#endif
+	    if (s1) {
+		close(*s1);
+		*s1 = -1;
+	    }
+	    if (s2) {
+		close(*s2);
+		*s2 = -1;
+	    }
+	    continue;
+	}
+		
+	if (ismc(res1->ai_addr)) {
+	    if (ismc(res2->ai_addr)) {
+		err("Both addresses cannot be multicast addresses");
+		return -1;
+	    }
+	    memcpy(mcsa, res1->ai_addr, res1->ai_addrlen);
+	    memcpy(ucsa, res2->ai_addr, res2->ai_addrlen);
+	    return 0;
+	}
+	if (ismc(res2->ai_addr)) {
+	    memcpy(ucsa, res1->ai_addr, res1->ai_addrlen);
+	    memcpy(mcsa, res2->ai_addr, res2->ai_addrlen);
+	    return 0;
+	}
+	err("Both addresses cannot be unicast addresses");
+	return -1;
     }
     return -1;
 }
-	    
-void parseargs(int argc, char **argv, int mode, char **name, int *family, uint32_t *intface,
-	       int *count, char **group) {
+
+void parseargs(int argc, char **argv, int mode, int *family, int *ver, uint16_t *size, uint32_t *intface,
+	       int *count, char **addr1, char **addr2, uint16_t *runtime, uint16_t *rate, char **srv) {
     int c;
     
     *family = AF_UNSPEC;
     *intface = 0;
     *count = 0;
-    *group = NULL;
+    *addr2 = NULL;
+    *ver = 0;
+    if (size)
+	*size = 0;
+    if (runtime)
+	*runtime = 0;
+    if (rate)
+	*rate = 0;
     
     while ((c = getopt(argc, argv,
 #ifdef WIN32
-		       "46c:"
+		       mode == FIRSTMODE ? "46vrc:t:" : "46vc:s:"
 #else
-		       "46I:c:"
+		       mode == FIRSTMODE ? "46vrI:c:t:" : "46vI:c:s:"
 #endif		       
 		       )) != -1) {
         switch (c) {
@@ -396,6 +600,13 @@ void parseargs(int argc, char **argv, int mode, char **name, int *family, uint32
         case '6':
 	    *family = AF_INET6;
 	    break;
+        case 'v':
+	    *ver = 1;
+	    break;
+        case 'r':
+	    if (rate)
+		*rate = 1;
+	    break;
 #ifndef WIN32
 	case 'I':
 	    *intface = if_nametoindex(optarg);
@@ -410,36 +621,86 @@ void parseargs(int argc, char **argv, int mode, char **name, int *family, uint32
 		break;
 	    fprintf(stderr, "Count must be positive\n");
 	    goto usage;
+	case 's':
+	    *size = atoi(optarg);
+	    if (*size > 0)
+		break;
+	    fprintf(stderr, "Invalid size\n");
+	    goto usage;
+	case 't':
+	    *runtime = atoi(optarg);
+	    if (*runtime > 0)
+		break;
+	    fprintf(stderr, "Invalid time to run\n");
+	    goto usage;
+
 	default:
 	    goto usage;
 	}
     }
 
-    if (mode == ASMMODE) {
+    switch (mode) {
+    case FIRSTMODE:
+	/* require two-three more arguments */
+	switch (argc - optind) {
+	case 2:
+	    *addr1 = argv[optind++];
+	    *srv = argv[optind];
+	    return;
+	case 3:
+	    *addr1 = argv[optind++];
+	    *addr2 = argv[optind++];
+	    *srv = argv[optind];
+	    return;
+	default:
+	    goto usage;
+	}
+    case ASMMODE:
 	/* require exactly two arguments after the options */
 	if (argc - optind != 2)
 	    goto usage;
-	*group = argv[optind++];
+	*addr2 = argv[optind++];
+	*addr1 = argv[optind];
+	return;
+    case SSMMODE:
+	/* require exactly one more argument */
+	if (argc - optind != 1)
+	    goto usage;
+	*addr1 = argv[optind];
+	return;
     }
-    /* require exactly one more argument */
-    if (argc - optind != 1)
-	goto usage;
-
-    *name = argv[optind];
-    return;
 
  usage:
-    fprintf(stderr,
+    fprintf(stderr, "%s version %s\n\n", argv[0], SSMPING_VERSIONSTRING);
+    switch (mode) {
+    case FIRSTMODE:
+	fprintf(stderr,
 #ifdef WIN32
-	    mode == SSMMODE
-	    ? "Usage:\n%s [ -46 ] [ -c count ] destination\n"
-	    : "Usage:\n%s [ -46 ] [ -c count ] group destination\n"
+		"Usage:\n%s [ -46vr ] [ -c count ] [ -t time ] [source] group port\nor\n%s [ -46vr ] [ -c count ] [ -t time ] group [source] port\n"
+#else    
+		"Usage:\n%s [ -46vr ] [ -I interface ] [ -c count ] [ -t time ] [source] group port\nor\n%s [ -46vr ] [ -I interface ] [ -c count ] [ -t time ] group [source] port\n"
+#endif    
+		, argv[0], argv[0]);
+	break;
+    case ASMMODE:
+	fprintf(stderr,
+#ifdef WIN32
+		"Usage:\n%s [ -46v ] [ -c count ] [ -s size ] group destination\nor\n%s [ -46v ] [ -c count ] [ -s size ] destination group\n"
+#else		
+		"Usage:\n%s [ -46v ] [ -I interface ] [ -c count ] [ -s size ] group destination\nor\n%s [ -46v ] [ -I interface ] [ -c count ] [ -s size ] destination group\n"
+#endif
+		, argv[0], argv[0]);
+	break;
+    case SSMMODE:
+	fprintf(stderr,
+#ifdef WIN32
+		"Usage:\n%s [ -46v ] [ -c count ] [ -s size ] destination\n"
 #else
-	    mode == SSMMODE
-	    ? "Usage:\n%s [ -46 ] [ -I interface ] [ -c count ] destination\n"
-	    : "Usage:\n%s [ -46 ] [ -I interface ] [ -c count ] group destination\n"
+		"Usage:\n%s [ -46v ] [ -I interface ] [ -c count ] [ -s size ] destination\n"
 #endif	    
-	    , argv[0]);
+		, argv[0]);
+	break;
+    }
     exit(1);
 }
 
@@ -450,9 +711,9 @@ void interrupt(int signo) {
     finish = 1;
 }
 
-int doit(int count, int us, int ms, struct sockaddr_storage *ucaddr,
+int doit(int ver, int size, int count, int us, int ms, struct sockaddr_storage *ucaddr,
 	 struct sockaddr_storage *grpaddr, char *source) {
-    int s, cnt, max, ndesc;
+    int s, cnt, max, ndesc, gotver = 0;
     int32_t hops;
     unsigned int uccount = 0, mccount = 0, mcminseq;
     long rtt, ucmin, ucmax, mcmin, mcmax, mdev;
@@ -464,9 +725,10 @@ int doit(int count, int us, int ms, struct sockaddr_storage *ucaddr,
     socklen_t fromlen = sizeof(from);
     size_t sendbuflen, recvbuflen, sendlen;
     struct timeval now, jointime, tstamp, diff, timeout, nextping;
-    char sendbuf[1024], recvbuf[1024];
-    void *sendseq, *sendtimestamp;
+    char mc, sendbuf[1024], recvbuf[1024], *sendtimestamp, *verstring = NULL;
+    void *sendseq;
     uint32_t u32t, seq = 0;
+    int32_t dist; 
 #ifndef WIN32
     struct sigaction sa_int;
 #endif
@@ -478,7 +740,7 @@ int doit(int count, int us, int ms, struct sockaddr_storage *ucaddr,
     
     pid = getpid();
     
-    sendlen = initsendbuf(sendbuf, sendbuflen, pid, grpaddr, &sendseq, &sendtimestamp);
+    sendlen = initsendbuf(sendbuf, sendbuflen, pid, ver, size, grpaddr, &sendseq, (void **)&sendtimestamp);
     
     max = us > ms ? us : ms;
 
@@ -530,7 +792,7 @@ int doit(int count, int us, int ms, struct sockaddr_storage *ucaddr,
 	s = FD_ISSET(us, &readfds) ? us : ms;
 
 	cnt = recvfromhopstime(s, (void *)&recvbuf, recvbuflen, 0,
-			   (struct sockaddr *)&from, &fromlen, &hops, &tstamp);
+			   (struct sockaddr *)&from, &fromlen, &hops, &tstamp, &mc);
 
 	if (cnt == -1) {
 	    err("recv failed");
@@ -552,7 +814,7 @@ int doit(int count, int us, int ms, struct sockaddr_storage *ucaddr,
 	    continue;
 	}
 	
-	parsepacket(recvbuf + 1, cnt - 1, &recvdata);
+	parsepacket(recvbuf + 1, cnt - 1, &verstring, &recvdata);
 	
 	if (recvdata.pid != pid) {
 	    printf("received someone else's reply\n");
@@ -561,23 +823,33 @@ int doit(int count, int us, int ms, struct sockaddr_storage *ucaddr,
 
 	timediff(&diff, &recvdata.timestamp, tstamp.tv_sec ? &tstamp : &now);
 
-#ifdef WIN32
-	printf("%s from %s, seq=%d time%c%ld ms",
-               s == us ? "  unicast" : "multicast",
-               addr2string((struct sockaddr *) &from, fromlen),
-	       recvdata.seq,
-	       (diff.tv_usec || diff.tv_sec) ? '=' : '<',
-	       (diff.tv_usec || diff.tv_sec) ? diff.tv_sec * 1000 + diff.tv_usec / 1000 : 1);
-#else
-	printf("%s from %s, seq=%d dist=%d time=%ld.%03ld ms",
-	       s == us ? "  unicast" : "multicast",
-	       addr2string((struct sockaddr *) &from, fromlen),
-	       recvdata.seq, hops == -1 ? -1 : 64 - hops,
-	       diff.tv_sec * 1000 + diff.tv_usec / 1000, diff.tv_usec % 1000);
+	if (verstring && !gotver) {
+	    gotver = 1;
+	    printf("Server version: %s\n", verstring);
+	}
+
+	if (size && cnt != size)
+	    printf("Warning, requested packet size %d, got %d\n", size, cnt);
+	
+	if (hops < 0)
+	    dist = -1;
+	else if (hops > 64) /* assuming Windows using ttl 128 */
+	    dist = 128 - hops;
+	else
+	    dist = 64 - hops;
+#ifndef WIN32
+	mc = s == ms;
 #endif
+	
+	printf("%s from %s, seq=%d dist=%d time=%ld.%03ld ms",
+	       mc ? "multicast" : "  unicast",
+	       addr2string((struct sockaddr *) &from, fromlen),
+	       recvdata.seq, dist,
+	       diff.tv_sec * 1000 + diff.tv_usec / 1000, diff.tv_usec % 1000);
+
 	rtt = (long)diff.tv_sec * 1000000 + (long)diff.tv_usec;
 
-	if (dupcheck(s == us, recvdata.seq)) {
+	if (dupcheck(mc, recvdata.seq)) {
 	    printf(" (DUP!)\n");
 	    continue;
 	}
@@ -619,7 +891,7 @@ int doit(int count, int us, int ms, struct sockaddr_storage *ucaddr,
     timediff(&diff, &jointime, &now);
     
     printf("\n");
-    printf("--- %s ssmping statistics ---\n", source);
+    printf("--- %s statistics ---\n", source);
     printf("%d packets transmitted, time %ld ms\n", seq, diff.tv_sec * 1000 + diff.tv_usec / 1000);
     printf("unicast:\n");
     printf("   %d packets received, %d%% packet loss\n", uccount, 100 * (seq - uccount) / seq);
